/*
 * Main File including the main-routine with initializations-
 * and stack-routines.
 *
 * Author: Simon Kueppers
 * Email: simon.kueppers@web.de
 * Homepage: http://klinkerstein.m-faq.de
 * */

#include <stdio.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include "Hardware/Enc28j60.h"
#include "Hardware/Spi.h"
#include "Hardware/Uart.h"
#include "Net/uip/uip.h"
#include "Net/uip/uip_arp.h"
#include "Net/uip/uip_TcpAppHub.h"
#include "Net/uip/uip_UdpAppHub.h"
#include <avr/eeprom.h>
#include <util/delay.h>
#include <string.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include "./Net/TcpApps/telnetd.h"
#include "Settings.h"
#include "Screens.h"
#include "Options.h"
#include "shell.h"

#define UART_BAUD_RATE      9600

char g_Buffer[128];

static uint16_t g_nPrescaler = 1000;
static volatile struct
{
	uint8_t fPeriodic :1;
} g_nFlags =
{ 0 };

static struct telnetd_state _s;
static struct telnetd_state* s = &_s;

//void uartprint(const char * _Format, ...)
void uartprint_P(const prog_char * _Format, ...)
{
	va_list argptr;
	va_start(argptr, _Format);
	//vsnprintf(s->buf, sizeof(s->buf), _Format, argptr);
	vsnprintf_P(s->buf, sizeof(s->buf), _Format, argptr);
	va_end(argptr);

	uart_puts(s->buf);
	s->bufptr = 0;
}

int main()
{
	unsigned int ret;
	uint8_t Return = 0;

	SpiInit();
	uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );
	Enc28j60Init();
	uip_arp_init();
	uip_init();
	uip_UdpAppHubInit();
	uip_TcpAppHubInit();
	Enc28j60SetClockPrescaler(2);

	//TCCR1B = (1<<WGM12)|(3<<CS10);
	//OCR1A = 25000; //16000000 / 64 / 25000 = 10Hz (100ms)
    //TIMSK1 |= (1 << OCIE1A);

	TCCR0A = (1<<WGM01);
	TCCR0B = (5<<CS00);
	OCR0A = 156; //16000000 / 1024 / 156 = 100Hz (10ms)
	TIMSK0 |= (1 << OCIE0A);

	// Config NIC
    if(eeprom_read_byte((uint8_t*)EE_DC))
    {
    	dhcpc_request();
    }
    else
    {
    	uip_ipaddr_t IpAddr;
    	eeprom_read_IP4(EE_IP, &IpAddr);
		uip_sethostaddr(IpAddr);

		eeprom_read_IP4(EE_NM, &IpAddr);
		uip_setnetmask(IpAddr);

		eeprom_read_IP4(EE_GW, &IpAddr);
		uip_setdraddr(IpAddr);;
    }

	sei	();

	// Laod Settings
	Load();

	// Init LCD
	InitScreen();

	// init serial terminal
    int i;
	for (i = 0; i < TELNETD_CONF_NUMLINES; ++i) {
		s->lines[i] = NULL;
	}
	s->bufptr = 0;
	s->state = STATE_NORMAL;
	shell_start(s);

	_delay_ms(100);

	// let the watchdog loose
    wdt_enable(WDTO_2S);

	while (1)
	{
		wdt_reset();

		uip_len = Enc28j60Receive(uip_buf);

		if (uip_len> 0)
		{
			if (((struct uip_eth_hdr *)&uip_buf[0])->type
					== htons(UIP_ETHTYPE_IP))
			{
				//uip_arp_ipin();
				uip_input();

				if (uip_len> 0)
				{
					uip_arp_out();
					Enc28j60CopyPacket(uip_buf, uip_len);
					Enc28j60Send();
				}

			}
			else if (((struct uip_eth_hdr *)&uip_buf[0])->type
					== htons(UIP_ETHTYPE_ARP))
			{
				uip_arp_arpin();
				if (uip_len> 0)
				{
					Enc28j60CopyPacket(uip_buf, uip_len);
					Enc28j60Send();
				}
			}
		}

		if (g_nFlags.fPeriodic) //Every 0.01 seconds
		{
			cli();
			g_nFlags.fPeriodic = 0;
			sei();

			int i = UIP_CONNS;
			while (i)
			{
				i--;
				uip_periodic(i);
				if (uip_len> 0)
				{
					uip_arp_out();
					Enc28j60CopyPacket(uip_buf, uip_len);
					Enc28j60Send();
				}
			}

#if UIP_UDP
			int j = UIP_UDP_CONNS;
			while (j)
			{
				j--;
				uip_udp_periodic(j);
				if (uip_len> 0)
				{
					uip_arp_out();
					Enc28j60CopyPacket(uip_buf, uip_len);
					Enc28j60Send();
				}
			}
#endif /* UIP_UDP */

			g_nPrescaler--;

			if (g_nPrescaler%10 == 0) //Every 0.1 seconds
			{

				ret = uart_getc();
				if ( (ret & UART_NO_DATA) == 0 )
				{
					if ( ret & UART_FRAME_ERROR ){
						uart_puts_P("UART Frame Error\r\n");
					} else if ( ret & UART_OVERRUN_ERROR ){
						uart_puts_P("UART Overrun Error\r\n");
					} else if ( ret & UART_BUFFER_OVERFLOW ){
						uart_puts_P("Buffer overflow error\r\n");
					}

					// echo
					if((char)ret != '\r' && (char)ret != '\n')
						uart_putc((unsigned char)ret);
					else if((char)ret == '\r')
						Return = 1;

					newchar((char)ret, s);
				}

				if(s->lines[0])
				{
					if(Return){
						Return = 0;
						uart_puts_P("\r\n");
					}
					uart_puts(s->lines[0]);
					free(s->lines[0]);
					for (i = 1; i < TELNETD_CONF_NUMLINES; ++i) {
						s->lines[i - 1] = s->lines[i];
					}
					s->lines[TELNETD_CONF_NUMLINES - 1] = NULL;
				}
			}

			CheckSwitch();
			if (g_nPrescaler%50 == 0) //Every 0.5 seconds
			{
				SwitchPage();
			}

			if (g_nPrescaler%100 == 0) //Every 1 seconds
			{
				MgmtDo();
				dhcpc_timer();
			}

			if (g_nPrescaler == 0) //Every 10 seconds
			{
				uip_arp_timer();
				g_nPrescaler = 1000;
			}
		}
	}

	return 0;
}

//ISR(TIMER1_COMPA_vect)
ISR(TIMER0_COMPA_vect)
{
	g_nFlags.fPeriodic = 1;
}
